/*  The code is written by R. Savinkov et al. while working on the article
    "Graph theory for modelling and analysis of the human lymphatic system",
   MDPI Mathematics. Feel free to contact by e-mail: r.savinkov@inm.ras.ru
    Please, while reuse or copy, save this information. */

#include <cmath>
#include <ctime>
#include <fstream>
#include <iostream>
#include <omp.h>
#include <set>
#include <string>
#include <vector>

using namespace std;

int main(int argc, char **argv) {
  int numNodes, numLayers, numFree;
  double reserveBranchProbability, overLayerBranchProbability;
  bool flag_numNodes = false, flag_numLayers = false, flag_numFree = false,
       flag_reserveBranchProbability = false,
       flag_overLayerBranchProbability = false;
  bool flag_noRulesForLayersWidth = false;
  int flagId = 1;
  setlocale(LC_ALL, "Russian");
  int nfl = 1;
  bool flag_draw = true;
  while (flagId < argc) {
    // with this, you can specify 2 or more output nodes
    string currentFlag = argv[flagId];
    if (currentFlag == "-nfl") {
      string currentValue = argv[flagId + 1];
      nfl = atoi(currentValue.c_str());
    }
    if (currentFlag == "-nr") {
      flag_noRulesForLayersWidth = true;
      cout << "Вы указали флаг \"-nr\". В данном режиме стабильность работы "
              "генератора не гарантируется."
           << endl;
      cout << "You declared flag \"-nr\". No any guarantees for working."
           << endl;
    }
    if (currentFlag == "-ndraw") {
      flag_draw = false;
    }
    if (currentFlag == "-nn") {
      string currentValue = argv[flagId + 1];
      numNodes = atoi(currentValue.c_str());
      flag_numNodes = true;
      flagId++;
    }
    if (currentFlag == "-nl") {
      string currentValue = argv[flagId + 1];
      numLayers = atoi(currentValue.c_str());
      flag_numLayers = true;
      flagId++;
    }
    if (currentFlag == "-nf") {
      string currentValue = argv[flagId + 1];
      numFree = atoi(currentValue.c_str());
      flag_numFree = true;
      flagId++;
    }
    if (currentFlag == "-rbp") {
      string currentValue = argv[flagId + 1];
      reserveBranchProbability = atof(currentValue.c_str());
      flag_reserveBranchProbability = true;
      flagId++;
    }
    if (currentFlag == "-olbp") {
      string currentValue = argv[flagId + 1];
      overLayerBranchProbability = atof(currentValue.c_str());
      flag_overLayerBranchProbability = true;
      flagId++;
    }
    flagId++;
  }
  srand(static_cast<unsigned int>(time(nullptr)));
  if (!flag_numNodes) {
    cout << "Укажите число узлов в графе:" << endl;
    cout << "Set nodes number: " << flush;
    cin >> numNodes;
  }
  while (numNodes < 2) {
    cout << "Внимание! Число узлов должно быть 2 или более. Укажите корректное "
            "число узлов:"
         << endl;
    cout << "Attention! The number of nodes must be 2 or more. Specify the "
            "appropriate number of nodes: "
         << flush;
    cin >> numNodes;
  }
  if (!flag_numLayers) {
    cout << "Укажите число слоёв в графе:" << endl;
    cout << "Specify layers number in graph: " << flush;
    cin >> numLayers;
  }
  while (numLayers < 2 || numLayers > numNodes) {
    cout << "Внимание! Число слоёв должно быть 2 или более, но не больше "
         << numNodes << ". Укажите корректное число слоёв:" << endl;
    cout << "Attention! The number of layers must be 2 or more, but not more "
            "than "
         << numNodes
         << ". Please enter the correct number of layers: " << flush;
    cin >> numLayers;
  }
  if (!flag_numFree) {
    cout << "Укажите число входов в графе:" << endl;
    cout << "Specify the number of entries in the graph: " << flush;
    cin >> numFree;
  }
  while (numFree < 1 || numFree > (numNodes - numLayers + 1)) {
    cout << "Внимание! Число входов должно быть 1 или более, но не больше "
         << numNodes - numLayers + 1
         << ". Укажите корректное число входов:" << endl;
    cout << "Attention! The number of inputs must be 1 or more, but not more "
            "than "
         << numNodes - numLayers + 1
         << ". Please enter the correct number of entries: " << flush;
    cin >> numFree;
  }
  if (!flag_reserveBranchProbability) {
    cout << "Укажите вероятность формирования резервных связей:" << endl;
    cout << "Indicate the likelihood of formation of additional edges: "
         << flush;
    cin >> reserveBranchProbability;
  }
  while (reserveBranchProbability < 0.0 || reserveBranchProbability >= 1.0) {
    cout << "Внимание! Вероятность не может быть отрицательным значением, а "
            "также она должна быть меньше 1. Укажите корректную вероятность: "
         << endl;
    cout << "Attention! The probability cannot be negative, and it must also "
            "be less than 1. Indicate the correct probability: "
         << flush;
    cin >> reserveBranchProbability;
  }
  if (!flag_overLayerBranchProbability) {
    cout << "Укажите вероятность формирования связи через слой:" << endl;
    cout << "Indicate the likelihood of forming an edge over the layer: "
         << flush;
    cin >> overLayerBranchProbability;
  }
  while (overLayerBranchProbability < 0.0) {
    cout << "Внимание! Вероятность не может быть отрицательным значением. "
            "Укажите корректную вероятность: "
         << flush;
    cout << "Attention! The probability cannot be negative. Enter the correct "
            "probability: "
         << flush;
    cin >> overLayerBranchProbability;
  }
  int *numInLine = new int[numLayers];
  int *layersRep = new int[numLayers];
  numInLine[0] = nfl;
  if (numNodes == numLayers)
    for (int i = 0; i < numLayers; i++)
      numInLine[i] = 1;
  else {
    int sumNodes = numLayers - 1 + nfl;
    for (int i = 1; i < numLayers; i++)
      numInLine[i] = 1;
    auto p_increaseNumNodesInLevels =
        [](int *numInLine, const int numNodes, const int numLayers,
           const int numFree, const bool flag_noRulesForLayersWidth) {
          const int aver_layer = (numNodes - numInLine[0]) / (numLayers - 1);
          int watchLayer = 1 + rand() % (numLayers - 1);
          if (!flag_noRulesForLayersWidth) {
            bool rulesExecuted =
                (watchLayer < (numLayers - 1) &&
                 numInLine[watchLayer + 1] > numInLine[watchLayer]) ||
                (watchLayer == (numLayers - 1) &&
                 numInLine[watchLayer] < (aver_layer * 1.1) &&
                 numInLine[watchLayer] < numFree);
            while (!rulesExecuted) {
              watchLayer = 1 + rand() % (numLayers - 1);
              rulesExecuted =
                  (watchLayer < (numLayers - 1) &&
                   numInLine[watchLayer + 1] > numInLine[watchLayer]) ||
                  (watchLayer == (numLayers - 1) &&
                   numInLine[watchLayer] < (aver_layer * 1.1) &&
                   numInLine[watchLayer] < numFree);
            }
          }
          numInLine[watchLayer]++;
        };
    while (sumNodes++ < numNodes)
      p_increaseNumNodesInLevels(numInLine, numNodes, numLayers, numFree,
                                 flag_noRulesForLayersWidth);
  }
  for (int i = 0; i < numLayers; i++)
    layersRep[i] = ((i == 0) ? numInLine[0] : numInLine[i] + layersRep[i - 1]);
  int maxLayer = 0;
  for (int i = 1; i < numLayers; i++)
    maxLayer = (numInLine[maxLayer] < numInLine[i] ? i : maxLayer);
  cout << "Предлагаемое разбиение по слоям:" << endl;
  cout << "Suggested layering:" << endl;
  for (int i = 0; i < numLayers; i++)
    cout << string((numInLine[maxLayer] - numInLine[i]) / 2, ' ') +
                string(numInLine[i], '*')
         << endl;
  bool **connectMatrix = new bool *[numNodes];
  for (int i = 0; i < numNodes; i++) {
    connectMatrix[i] = new bool[numNodes];
    for (int j = 0; j < numNodes; j++)
      connectMatrix[i][j] = false;
  }
  cout << "Аллокация матрицы выполнена." << endl;
  cout << "Matrix allocation done." << endl;
  auto f_countFreeNodes = [](bool **connectMatrix, int numNodes) {
    int r = 0;
    for (int i = 0; i < numNodes; i++) {
      int counter = 0;
      for (int j = 0; j < numNodes; j++)
        counter += ((j > i && connectMatrix[i][j]) ? 1 : 0);
      counter ? 0 : r++;
    }
    return r;
  };
  auto f_countDisconnectedNodes = [](bool **connectMatrix, int numNodes) {
    int r = 0;
    for (int i = 0; i < numNodes; i++) {
      int counter = 0;
      for (int j = 0; j < numNodes; j++)
        counter += ((j < i && connectMatrix[i][j]) ? 1 : 0);
      counter ? 0 : r++;
    }
    return r;
  };
  cout << "Формирование связей." << endl;
  cout << "Forming connections." << endl;
  set<int> inputsList;
  vector<int> freeVertices;
  for (int i = 0; i < numLayers; i++)
    freeVertices.push_back(numInLine[i]);
  for (int i = layersRep[numLayers - 1] - numInLine[numLayers - 1];
       i < layersRep[numLayers - 1]; i++)
    inputsList.insert(i);
  while (inputsList.size() < numFree) {
    int ptr = nfl + rand() % (numNodes - nfl);
    int nodeLayer = 0;
    while (ptr >= layersRep[nodeLayer])
      nodeLayer++;
    if (inputsList.find(ptr) == inputsList.end() &&
        freeVertices[nodeLayer] > 1) {
      freeVertices[nodeLayer]--;
      inputsList.insert(ptr);
    }
  }
  for (int i = nfl; i < numNodes; i++) {
    cout << "[" << i << "]: ";
    int nodeLayer = 0;
    while (i >= layersRep[nodeLayer])
      nodeLayer++;
    cout << nodeLayer << endl;
    int createReferenceTo;
    vector<int> verticesWithNoInputs, possibleForConnectionVertices;
    for (int i = layersRep[nodeLayer - 1] - numInLine[nodeLayer - 1];
         i < layersRep[nodeLayer - 1]; i++) {
      bool no_inputs = true;
      for (int j = i + 1; j < numNodes; j++)
        if (connectMatrix[j][i] || connectMatrix[i][j])
          no_inputs = false;
      if (no_inputs && inputsList.find(i) == inputsList.end())
        verticesWithNoInputs.push_back(i);
      if (inputsList.find(i) == inputsList.end())
        possibleForConnectionVertices.push_back(i);
    }
    if (verticesWithNoInputs.size())
      createReferenceTo =
          verticesWithNoInputs[rand() % verticesWithNoInputs.size()];
    else
      createReferenceTo =
          possibleForConnectionVertices[rand() %
                                        possibleForConnectionVertices.size()];
    connectMatrix[i][createReferenceTo] = connectMatrix[createReferenceTo][i] =
        true;
    if (inputsList.find(i) == inputsList.end())
      while (((reserveBranchProbability * RAND_MAX) > rand())) {
        int connectingWithLayer = nodeLayer - 1;
        while (connectingWithLayer > 0 &&
               ((overLayerBranchProbability * RAND_MAX) > rand()))
          connectingWithLayer--;
        possibleForConnectionVertices.clear();
        for (int i = layersRep[connectingWithLayer] -
                     numInLine[connectingWithLayer];
             i < layersRep[connectingWithLayer]; i++)
          if (inputsList.find(i) == inputsList.end())
            possibleForConnectionVertices.push_back(i);
        if (possibleForConnectionVertices.size()) {
          int newTargetForBranch = possibleForConnectionVertices
              [rand() % possibleForConnectionVertices.size()];
          connectMatrix[i][newTargetForBranch] =
              connectMatrix[newTargetForBranch][i] = true;
        }
      }
  }
  cout << "Формирование связей выполнено. Коррекция свободных вершин." << endl;
  cout << "Link formation is complete. Free vertex correction." << endl;
  cout << "Free vertices: " << f_countFreeNodes(connectMatrix, numNodes)
       << endl;
  auto p_tryMakeNodeNonfree = [](const int nfl, int *layersRep,
                                 bool **connectMatrix, int *numInLine,
                                 int numLayers, int numNodes) {
    cout << "Entering p_tryMakeNodeNonfree-lambda function" << endl;
    bool success = false;
    int nodeId, nodeLayer;
    while (!success) {
      nodeId = nfl + rand() % (numNodes - nfl - numInLine[numLayers - 1]);
      nodeLayer = 0;
      while (nodeId >= layersRep[nodeLayer]) {
        nodeLayer++;
      }
      int nodeLineStartFrom = layersRep[nodeLayer] - numInLine[nodeLayer];
      int upperLineStartFrom =
          layersRep[nodeLayer + 1] - numInLine[nodeLayer + 1];
      int candidates = 0;
      bool nodeIsFree = true;
      for (int i = nodeLineStartFrom; i < layersRep[nodeLayer]; i++) {
        int haveIncoming = 0;
        if (i != nodeId) {
          for (int j = i + 1; j < numNodes; j++) {
            if (connectMatrix[i][j]) {
              haveIncoming++;
            }
          }
        } else {
          for (int j = i + 1; j < numNodes; j++) {
            if (connectMatrix[i][j]) {
              nodeIsFree = false;
            }
          }
        }
        haveIncoming > 1 ? candidates++ : 0;
      }
      if (nodeIsFree && candidates > 0) {
        success = true;
        int *candidatesList = new int[candidates];
        candidates = 0;
        for (int i = nodeLineStartFrom; i < layersRep[nodeLayer]; i++) {
          int haveIncoming = 0;
          for (int j = i + 1; j < numNodes; j++) {
            if (i != nodeId && connectMatrix[i][j]) {
              haveIncoming++;
            }
          }
          if (haveIncoming > 1) {
            candidatesList[candidates++] = i;
          }
        }
        int useCandidate = candidatesList[rand() % candidates];
        delete[] candidatesList;
        int candidateConnections = 0;
        for (int i = useCandidate + 1; i < numNodes; i++) {
          connectMatrix[i][useCandidate] ? candidateConnections++ : 0;
        }
        candidatesList = new int[candidateConnections];
        candidateConnections = 0;
        for (int i = useCandidate + 1; i < numNodes; i++) {
          if (connectMatrix[i][useCandidate]) {
            candidatesList[candidateConnections++] = i;
          }
        }
        int upperLayerCandidate = candidatesList[rand() % candidateConnections];
        delete[] candidatesList;
        connectMatrix[useCandidate][upperLayerCandidate] =
            connectMatrix[upperLayerCandidate][useCandidate] = false;
        connectMatrix[nodeId][upperLayerCandidate] =
            connectMatrix[upperLayerCandidate][nodeId] = true;
      }
    }
    cout << "Leaving p_tryMakeNodeNonfree-lambda function" << endl;
  };
  if (f_countFreeNodes(connectMatrix, numNodes) > numFree) {
    cout << "Число свободных вершин больше требуемого. Корректировка." << endl;
    cout << "The number of free vertices is more than required. Correction."
         << endl;
  }
  if (f_countFreeNodes(connectMatrix, numNodes) < numFree) {
    cout << "Число свободных вершин меньше требуемого. Корректировка." << endl;
    cout << "The number of free vertices is less than required. Correction."
         << endl;
  }
  while (f_countFreeNodes(connectMatrix, numNodes) > numFree) {
    p_tryMakeNodeNonfree(nfl, layersRep, connectMatrix, numInLine, numLayers,
                         numNodes);
  }
  for (int i = 0; i < numLayers; i++) {
    cout << (i ? " " : "") << layersRep[i];
  }
  cout << endl;
  while (f_countFreeNodes(connectMatrix, numNodes) < numFree) {
    auto p_tryMakeNodeFree = [](const int nfl, int *layersRep,
                                bool **connectMatrix, int *numInLine,
                                int numLayers, int numNodes) {
      cout << "Entering p_tryMakeNodeFree-lambda function" << endl;
      bool success = false;
      int nodeId, nodeLayer;
      while (!success) {
        cout << "\r    \r"
             << "#" << flush;
        nodeId = nfl + rand() % (numNodes - nfl - numInLine[numLayers - 1]);
        nodeLayer = 0;
        while (nodeId >= layersRep[nodeLayer]) {
          nodeLayer++;
        }
        int nodeLineStartFrom = layersRep[nodeLayer] - numInLine[nodeLayer];
        int upperLineStartFrom =
            layersRep[nodeLayer + 1] - numInLine[nodeLayer + 1];
        int candidates = 0;
        bool nodeIsNonfree = false;
        cout << "#" << flush;
        for (int i = nodeLineStartFrom; i < layersRep[nodeLayer]; i++) {
          bool haveIncoming = false;
          if (i != nodeId) {
            for (int j = i + 1; j < numNodes; j++) {
              if (connectMatrix[i][j]) {
                haveIncoming = true;
              }
            }
          } else {
            for (int j = upperLineStartFrom; j < layersRep[nodeLayer + 1];
                 j++) {
              if (connectMatrix[i][j]) {
                nodeIsNonfree = true;
              }
            }
          }
          haveIncoming ? candidates++ : 0;
        }
        cout << "#" << flush;
        if (nodeIsNonfree && candidates > 0) {
          cout << "+" << flush;
          success = true;
          int *candidatesList = new int[candidates];
          candidates = 0;
          for (int i = nodeLineStartFrom; i < layersRep[nodeLayer]; i++) {
            bool haveIncoming = false;
            for (int j = i + 1; j < numNodes; j++) {
              if (i != nodeId && connectMatrix[i][j]) {
                haveIncoming = true;
              }
            }
            if (haveIncoming) {
              candidatesList[candidates++] = i;
            }
          }
          for (int i = upperLineStartFrom; i < layersRep[numLayers - 1]; i++) {
            if (connectMatrix[i][nodeId]) {
              connectMatrix[i][nodeId] = connectMatrix[nodeId][i] = false;
              int targetNodeId = candidatesList[rand() % candidates];
              connectMatrix[i][targetNodeId] = connectMatrix[targetNodeId][i] =
                  true;
            }
          }
          delete[] candidatesList;
        }
      }
      cout << "Leaving p_tryMakeNodeFree-lambda function" << endl;
    };
    p_tryMakeNodeFree(nfl, layersRep, connectMatrix, numInLine, numLayers,
                      numNodes);
  }
  ofstream fout("graph.gv", ofstream::out);
  if (fout.is_open()) {
    fout << "digraph the_graph {" << endl;
    for (int i = 0; i < numNodes; i++) {
      for (int j = i + 1; j < numNodes; j++) {
        if (connectMatrix[i][j]) {
          fout << "    n" << j << " -> n" << i << ";" << endl;
        }
      }
    }
    fout << "}";
    fout.close();
    if (flag_draw) {
      cout << "Вызываем отрисовку графа (drawing graph):\n    dot -Tpng "
              "graph.gv -o "
              "graph.png\n        --> "
           << flush
           << (system("dot -Tpng graph.gv -o graph.png")
                   ? "Что-то пошло не так (something is wrong)."
                   : "Успешно (all is fine).")
           << endl;
    }
  }
  fout.open("gnuplot_matrix.txt", ofstream::out);
  if (fout.is_open()) {
    fout << numNodes;
    for (int i = 0; i < numNodes; i++) {
      fout << " " << i;
    }
    fout << endl;
    for (int i = 0; i < numNodes; i++) {
      fout << i;
      for (int j = 0; j < numNodes; j++) {
        fout << " " << ((connectMatrix[i][j] && i > j) ? 1 : 0);
      }
      fout << endl;
    }
    fout.close();
  }
  fout.open("matrix.m", ofstream::out);
  if (fout.is_open()) {
    fout << "Matrix = [";
    for (int i = 0; i < numNodes; i++) {
      for (int j = 0; j < numNodes; j++) {
        fout << " " << (connectMatrix[i][j] ? (i > j ? 1 : 0) : 0);
      }
      fout << endl;
    }
    fout << "];";
    fout.close();
  }
  fout.open("matrix.txt", ofstream::out);
  if (fout.is_open()) {
    fout << numNodes << endl;
    for (int i = 0; i < numNodes; i++) {
      for (int j = 0; j < numNodes; j++) {
        fout << (j ? " " : "") << (connectMatrix[i][j] ? (i > j ? 1 : 0) : 0);
      }
      fout << endl;
    }
    fout.close();
  }
  fout.open("gnuplot_script.txt", ofstream::out);
  if (fout.is_open()) {
    fout << "set terminal png size 12800,10800 font \"Comic Sans MS, 144\""
         << endl;
    fout << "set autoscale xfix" << endl;
    fout << "set autoscale yfix" << endl;
    fout << "set palette defined ( 0 '#ffffff', \\" << endl;
    fout << "    1 '#000000' )" << endl;
    fout << "set size square" << endl;
    fout << "set output 'matrix.png'" << endl;
    fout << "plot 'gnuplot_matrix.txt' matrix nonuniform with image notitle";
    fout.close();
    if (flag_draw) {
      cout << "Вызываем отрисовку матрицы (drawing the graph matrix):\n    "
              "gnuplot gnuplot_script.txt\n        "
              "--> "
           << flush
           << (system("gnuplot gnuplot_script.txt")
                   ? "Что-то пошло не так (something is wrong)."
                   : "Успешно (all is fine).")
           << endl;
    }
  }
  return 0;
}
